@@ -23,6 +23,12 @@ class AppCenterHTTPError:
     code: str
     message: str
 
+    def __str__(self) -> str:
+        """Generate and return the string representation of the object.
+        :return: A string representation of the object
+        """
+        return f"<code={self.code}, message={self.message}>"
+
 
 class AppCenterHTTPException(Exception):
     """All App Center HTTP exceptions use this class."""
@@ -43,10 +49,10 @@ class AppCenterHTTPException(Exception):
         :return: A string representation of the object
         """
         return (
-            f"method={self.response.request.method}, "
+            f"<method={self.response.request.method}, "
             + f"url={self.response.request.url}, "
             + f"status_code={self.response.status_code}, "
-            + f"text={self.response.text}"
+            + f"text={self.response.text}>"
         )
 
 
@@ -127,6 +133,18 @@ class AppCenterDerivedClient:
         self.session = requests.Session()
         self.session.headers.update({"X-API-Token": self.token})
 
+    # pylint: disable=no-self-use
+    def base_url(self, *, version: str = "0.1") -> str:
+        """Generate the base URL for the API.
+
+        :param str version: The API version to hit
+
+        :returns: The base URL
+        """
+        return f"{API_BASE_URL}/v{version}"
+
+    # pylint: enable=no-self-use
+
     def generate_url(self, *, version: str = "0.1", owner_name: str, app_name: str) -> str:
         """Generate a URL to use for querying the API.
 
@@ -137,7 +155,7 @@ class AppCenterDerivedClient:
         :returns: A generated URL base
         """
 
-        url = f"{API_BASE_URL}/v{version}/apps/{owner_name}/{app_name}"
+        url = self.base_url(version=version) + f"/apps/{owner_name}/{app_name}"
         self.log.debug(f"Generated URL: {url}")
         return url
 
@@ -212,6 +230,28 @@ class AppCenterDerivedClient:
 
         return response
 
+    @retry(
+        retry=(retry_if_exception(_is_connection_failure)),
+        wait=wait_fixed(10),
+        stop=stop_after_attempt(3),
+    )
+    def post_raw_data(self, url: str, data: Any) -> requests.Response:
+        """Perform a POST request to a url
+
+        :param url: The URL to run the POST on
+        :param data: The JSON serializable data to send
+
+        :returns: The raw response
+
+        :raises AppCenterHTTPException: If the request fails with a non 200 status code
+        """
+        response = self.session.post(url, headers={}, data=data)
+
+        if response.status_code < 200 or response.status_code >= 300:
+            raise create_exception(response)
+
+        return response
+
     @retry(
         retry=(retry_if_exception(_is_connection_failure)),
         wait=wait_fixed(10),
